import datetime
import time

from injector import singleton

from src.fastinject import inject, get_default_registry, injectable


def test_injectable_servcie_is_injectable():
    @injectable()
    class TimeStamp:
        ts: float

        def __init__(self) -> None:
            self.ts = time.time()

        @property
        def datetime_str(self) -> str:
            return datetime.datetime.fromtimestamp(self.ts).strftime("%Y-%m-%d %H:%M:%S")

    @inject()
    def function_with_injection(ts: TimeStamp):
        assert ts is not None

    # set_default_registry(registry=None)
    function_with_injection()
    reg = get_default_registry()
    assert reg is not None


def test_injectable_service_is_not_singleton():
    @injectable()
    class _TimeStamp:
        ts: float

        def __init__(self) -> None:
            self.ts = time.time()

    @inject()
    def function_with_injection(ts: _TimeStamp):
        id_1 = id(ts)
        ts_1 = ts.ts

        @inject()
        def other_function_with_injection(ts: _TimeStamp):
            assert id(ts) != id_1
            assert ts.ts != ts_1

        time.sleep(0.1)
        other_function_with_injection()

    # set_default_registry(registry=None)
    reg = get_default_registry()
    assert reg is not None
    function_with_injection()


def test_injectable_servcie_singleton_works():
    @injectable(scope=singleton)
    class TimeStamp:
        ts: float

        def __init__(self) -> None:
            self.ts = time.time()

    @inject()
    def function_with_injection(ts: TimeStamp):
        id_1 = id(ts)
        ts_1 = ts.ts

        @inject()
        def other_function_with_injection(ts: TimeStamp):
            assert id(ts) == id_1
            assert ts.ts == ts_1

        time.sleep(0.1)
        other_function_with_injection()

    # set_default_registry(registry=None)
    function_with_injection()



import logging
import time

import pytest

from src.fastinject import (
    Registry,
)
from test.objects_for_testing import services
from test.objects_for_testing.service_configs import (
    SCLogging,
    SCDatabase,
    SCTimestamper,
    SCTimestamperWeirdImport,
)
from test.objects_for_testing.services import DatabaseConfig, TimeStamp


@pytest.fixture(scope="function")
def test_get_from_registry():
    """Test example"""
    registry = Registry(service_configs=[SCLogging])
    assert registry.get(logging.Logger) is not None
    assert registry.get(SCTimestamper) is None


@pytest.fixture(scope="function")
def test_can_get_from_registry_folder_import():
    """The service is imported like folder.Classname in the Module"""
    registry = Registry(service_configs=[SCTimestamperWeirdImport])
    assert registry.get(services.TimeStamp) is not None
    assert registry.get(TimeStamp) is not None


def test_get_from_registry_autobind():
    """Even though ModuleTimestamp is not registered on the registry, when we auto_bind=True, injector will resolve the dependency.
    This is becauase it searches the type in all functions decorated with @provider"""
    # Create registry
    registry = Registry(service_configs=[SCDatabase], auto_bind=True)

    # Can find both modules even though
    assert registry.get(DatabaseConfig) is not None
    assert registry.get(TimeStamp) is not None

    # Assert that ts is a full TimeStamp with all methods etc.
    ts: TimeStamp = registry.get(TimeStamp)
    assert ts is not None  # should be found
    assert "time_passed" in dir(ts)  # has 'time_passed' function
    assert isinstance(ts.time_passed(), float)
    time.sleep(0.001)
    assert ts.time_passed() > 0.0



from typing import get_origin, Union, get_args, Type


def is_optional_type(param_type: Type) -> bool:
    """Check if a type is Optional (Union with None).
    Type is Optional when it has a default value
    """
    origin = get_origin(param_type)
    return origin is Union and type(None) in get_args(param_type)


def get_type_that_optional_wraps(param_type: Type) -> Type:
    """Extract the actual type from an Optional type."""
    if is_optional_type(param_type):
        return next(arg for arg in get_args(param_type) if arg is not type(None))
    return param_type



import logging

import pytest
from injector import provider, singleton

from src.fastinject import Registry, set_default_registry
from src.fastinject.service_config import ServiceConfig
from test.objects_for_testing import services
from test.objects_for_testing.services import DatabaseConfig, TimeStamp


# @pytest.fixture(scope="function", autouse=True)
# def setup_function():
#     set_default_registry(registry=None)

@pytest.fixture(scope="function")
def test_singleton_returns_same_object():
    class SCTimestamper(ServiceConfig):
        @singleton
        @provider
        def provide_timestamper(self) -> services.TimeStamp:
            return services.TimeStamp()

    registry = Registry(service_configs=[SCTimestamper], auto_bind=True)

    # Can find both modules even though
    assert registry.get(TimeStamp) is not None


@pytest.fixture(scope="function")
def test_non_singleton_returns_different_object():
    class SCTimestamper(ServiceConfig):
        @provider
        def provide_timestamper(self) -> services.TimeStamp:
            return services.TimeStamp()

    registry = Registry(service_configs=[SCTimestamper], auto_bind=True)

    # Can find both modules even though
    assert registry.get(TimeStamp) is not None

    id1 = id(registry.get(TimeStamp))
    id2 = id(registry.get(TimeStamp))
    assert id1 != id2



import logging
from typing import Optional

import pytest

from src.fastinject import Registry, get_default_registry
from src.fastinject.decorators import inject
from test.objects_for_testing import services, service_configs
from test.objects_for_testing.service_configs import SCDatabaseLogging


def test_inject_from_default_registry():
    """Test example"""
    # 1. Create registry
    registy = Registry(service_configs=[SCDatabaseLogging])
    assert registy is not None
    assert get_default_registry() is not None

    @inject()
    def getlogger(_logger: logging.Logger) -> None:
        assert _logger is not None

    getlogger()


@pytest.fixture(scope="function")
def test_inject_raises_if_module_not_registered_from_default_registry_double():
    """Test example"""
    # 1. Create registry
    registy = Registry(
        service_configs=[
            SCDatabaseLogging,
        ]
    )  # ModuleTimestamper])
    assert registy is not None
    assert get_default_registry() is not None

    @inject()
    def getlogger(ts: services.TimeStamp, logger: logging.Logger) -> None: ...

    with pytest.raises(TypeError):
        getlogger()


def test_injects_none_on_missing_service():
    """Test example"""
    # 1. Create registry
    registy = Registry(service_configs=[SCDatabaseLogging])
    assert registy is not None
    assert get_default_registry() is not None

    @inject()
    def getlogger(_logger: logging.Logger, ts: Optional[services.TimeStamp] = None) -> None:
        assert _logger is not None
        assert ts is None

    getlogger()


def test_imperative():
    reg = get_default_registry()
    assert reg is not None


def test_raises_on_failed_service_init():
    """Test example"""
    # 1. Create registry
    registy = Registry(service_configs=[service_configs.SCDatabaseInitFails])
    assert registy is not None
    assert get_default_registry() is not None

    # this will init the services
    with pytest.raises(Exception):
        d = registy.get(services.DatabaseConnection)


# def test_inject_raises_if_no_registry_set():
#     """Injector searches in each function decorated with @provider"""
#     # 1. Make sure there is no global registry
#     # set_default_registry(None)
#     # registy = Registry(modules=[ModuleDatabaseLogging])
#
#     @inject()
#     def getlogger(ts:services.TimeStamp, logger: Optional[logging.Logger] = None) -> None:
#         ...
#
#     with pytest.raises(ValueError):
#         getlogger()



import inspect
from functools import wraps
from typing import Callable, Optional, Type

from injector import Scope

from .loggers import logger
from .registry import Registry, get_default_registry
from .service_config import ServiceConfig
from .type_helpers import is_optional_type, get_type_that_optional_wraps


def inject_from(registry: Optional[Registry] = None, inject_missing_optional_as_none: bool = True) -> Callable:
    """Decorator that inspects the decorated function and injects instances from the provided registry if available."""

    def decorator(func: Callable) -> Callable:
        # Get decorated function's signature
        fn_signature = inspect.signature(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            # Get decorated function's bound_arguments
            bound_arguments = fn_signature.bind_partial(*args, **kwargs)
            bound_arguments.apply_defaults()

            # Skip params that already have an argument
            # Check which params misses arguments for us to look up in the registry;
            for param_name, param_val in fn_signature.parameters.items():
                is_optional_param: bool = is_optional_type(param_type=param_val.annotation)
                if param_name in bound_arguments.arguments:
                    has_value: bool = (
                        param_name in bound_arguments.arguments and bound_arguments.arguments[param_name] is not None
                    )
                    value_is_none = bound_arguments.arguments[param_name] is None
                    if has_value or (is_optional_param and value_is_none):
                        logger.debug(
                            f"Skipping injection for '{param_name}' - already provided or optional with default None."
                        )
                        continue

                # Attempt to retrieve from registry
                type_to_resolve: Type = get_type_that_optional_wraps(param_val.annotation)

                found_service = None
                logger.debug(f"Resolving parameter '{param_name}':{param_val.annotation} (type={type_to_resolve})")
                try:
                    found_service = registry.get(type_to_resolve, None)
                    logger.debug(f"found service: {found_service}")
                except Exception as e:
                    caller_provided_value: bool = param_name in bound_arguments.arguments
                    if not caller_provided_value and not is_optional_param:
                        raise ValueError(f"Cannot inject value for required parameter '{param_name}'. Error: {e}")
                    # eigher value for param is provided by user or param is optional; keep found_service=None

                # Actually inject a value, albeit None or the found service
                if found_service is not None:
                    logger.debug(f"Injecting resolved value for the parameter '{param_name}'")
                    bound_arguments.arguments[param_name] = found_service
                elif inject_missing_optional_as_none and is_optional_param:
                    # inject None for optional values if it can't be resolved
                    logger.debug(f"Injecting NONE for the optional parameter '{param_name}'")
                    bound_arguments.arguments[param_name] = None

            return func(*bound_arguments.args, **bound_arguments.kwargs)

        return wrapper

    return decorator


def inject(inject_missing_optional_as_none: bool = True) -> Callable:
    """Decorator that inspects the decorated function and injects instances from the provided registry if available."""

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            target_registry = get_default_registry()
            if target_registry is None:
                raise ValueError("No registries configured")

            inject_func = inject_from(
                registry=target_registry, inject_missing_optional_as_none=inject_missing_optional_as_none
            )(func)
            return inject_func(*args, **kwargs)

        return wrapper

    return decorator


def injectable(scope: Optional[Scope] = None):
    def decorator(original_class: Type):
        registry = get_default_registry() or Registry()

        if issubclass(original_class, ServiceConfig):
            registry.add_service_config(service_config=original_class)
            return original_class

        # def configure_for_testing(binder: Binder):
        #     """ Puts a service in a registry without the decorators """
        #     binder.bind(original_class, to=original_class, scope=scope)
        # registry.add_setup_function(configure_for_testing)
        registry.add_service(service=original_class, scope=scope)

        return original_class

    return decorator


# def injectables(original_class: Type):
#
#     registry = get_default_registry() or Registry()
#
#     if not issubclass(original_class, ServiceConfig):
#         raise ValueError(f"{original_class} does not inherit from Module. Must inherit from Module to be injectable.")
#     registry.add_service_config(service_config=original_class)
#     return original_class



from typing import NewType

from src.fastinject import Registry, ServiceConfig, provider
from test.objects_for_testing import services


def test_can_register_multiple_instances_of_same_type():
    DBConfig1 = NewType("DatabaseConfig", services.DatabaseConfig)
    DBConfig2 = NewType("DatabaseConfig", services.DatabaseConfig)

    class MyServiceConfig(ServiceConfig):
        @provider
        def provide_config1(self) -> DBConfig1:
            return services.DatabaseConfig(connection_string="constring_1")

        @provider
        def provide_config2(self) -> DBConfig2:
            return services.DatabaseConfig(connection_string="constring_2")

    registry = Registry(service_configs=[MyServiceConfig])
    assert registry.get(DBConfig1).connection_string == "constring_1"
    assert registry.get(DBConfig2).connection_string == "constring_2"



# from src.injectr import singleton, Registry
#
# from test.objects_for_testing import services
#
#
# def test_can_register_service_on_registry_imperatively():
#     """ For binding a service to a module on the fly """
#     def configure_for_testing(binder):
#         """ Puts a service in a registry without the decorators """
#         configuration = services.MyDatabaseConfig("file:memdb1?mode=memory&cache=shared")
#         binder.bind(services.MyDatabaseConfig, to=configuration, scope=singleton)
#
#     # Create empty registry; no modules added
#     registry = Registry()
#     assert registry._modules == []
#     registry.add_setup_function(configure_for_testing)
#
#     service: services.MyDatabaseConfig = registry.get(services.MyDatabaseConfig)
#     assert service.connection_string == "file:memdb1?mode=memory&cache=shared"
#
#



import logging

from src.fastinject import Registry
from test.objects_for_testing import services
from test.objects_for_testing.service_configs import SCLogging, SCDatabase
from test.objects_for_testing.services import DatabaseConfig


def test_create_registry_init_works():
    """Test example"""
    registry_builder = Registry()
    assert len(registry_builder._modules) == 0
    registry_builder.add_service_config(SCLogging)
    assert len(registry_builder._modules) == 1
    assert registry_builder.get(logging.Logger) is not None
    assert len(registry_builder._modules) == 1


def test_create_registry_pass_modules_in_init_works():
    """Test example"""
    registry_builder = Registry(service_configs=[SCLogging, SCDatabase])
    assert len(registry_builder._modules) == 2
    assert registry_builder.get(logging.Logger) is not None


def test_create_registry_init_then_add_works():
    """Test example"""
    registry_builder = Registry(service_configs=[SCDatabase])
    assert len(registry_builder._modules) == 1
    registry_builder.add_service_config(service_config=SCLogging)
    assert len(registry_builder._modules) == 2
    assert registry_builder.get(logging.Logger) is not None
    assert registry_builder.get(DatabaseConfig) is not None


def test_create_registry_add_service_works():
    registy = Registry()
    registy.add_service(service=services.TimeStamp)
    assert registy.get(interface=services.TimeStamp) is not None


def test_create_registry_with_services():
    registy = Registry(services=[services.TimeStamp])
    assert registy.get(interface=services.TimeStamp) is not None



from typing import Any, Type, Union, Callable, List, Optional  # , Self,  Iterable

from injector import Injector, Binder
from injector import Scope, ScopeDecorator, Module
from injector import T
from .service_config import ServiceConfig


# _InstallableModuleType = Union[Callable[['XBinder'], None], 'Module', Type['Module']]
_InstallableModuleType = Union["Service", Type["Service"]]


# class XBinder(Binder):
#     def __init__(self, parent: Binder) -> None:
#         self._parent = parent
#
#     def bind(
#         self,
#         interface: Type[T],
#         to: Union[None, T, Callable[..., T], Provider[T]] = None,
#         scope: Union[None, Type["Scope"], "ScopeDecorator"] = None,
#     ) -> None:
#         self._parent.bind(interface, to, scope)


class Registry:
    """_setup_functions contain functions that let us imperatively register a service within a dependency injection Registry.
    It dynamically binds a configuration object to the registry without using decorators, allowing for on-the-fly service binding
    """

    _auto_bind: bool
    _modules: List[ServiceConfig]
    _setup_functions: List[Callable[[Binder], None]]

    def __init__(
        self,
        service_configs: Optional[List[_InstallableModuleType]] = None,
        services: Optional[List[_InstallableModuleType]] = None,
        auto_bind: bool = False,
    ) -> None:
        """
        auto_bind: bool : if True; will automatically resolve types that it can construct. Looks in all functions decorated with @provider.

        """
        self._auto_bind = auto_bind
        self._modules: List[ServiceConfig] = []
        self._setup_functions: List[Callable[[Binder], None]] = []
        for service_config in service_configs or []:
            self.add_service_config(service_config=service_config)
        for service in services or []:
            self.add_service(service=service)
        self.__build()

    def add_service(self, service: Type, scope: Optional[Scope] = None) -> "Registry":
        def _setup_function(binder: Binder):
            binder.bind(service, to=service, scope=scope)

        self._setup_functions.append(_setup_function)
        self.__build()
        return self

    def add_service_config(self, service_config: _InstallableModuleType) -> "Registry":
        self._modules.append(service_config)
        self.__build()
        return self

    # def bind(
    #     self,
    #     interface: Type[T],
    #     to: Union[None, T, Callable[..., T], Provider[T]] = None,
    #     scope: Union[None, Type["Scope"], "ScopeDecorator"] = None,
    # ) -> "T":
    #     return self.add_setup(lambda binder: binder.bind(interface, to, scope))

    def __build(self) -> "Registry":
        def configure(binder: Binder):
            for fn in self._setup_functions:
                fn(binder)

        # setup:ParentListList = [configure]
        setup: List[Callable] = [configure]
        for m in self._modules:
            setup.append(m)

        # self._injector = Injector([configure_for_testing, di.TestModule()])
        self._injector = Injector(setup, auto_bind=self._auto_bind)
        set_default_registry(registry=self)
        return self
        # result = RegistryDEPRECATED(self._injector)
        # for modifier in self._modifiers:
        #     # if self._modifier:
        #     modifier(result, self._injector)

        # auto register the first registry as default registry
        # if get_default_registry() is None:
        #     set_default_registry(result)
        # return result

    def get(self, interface: Type[T], scope: Union[ScopeDecorator, Type[Scope], None] = None) -> T:
        """
        Gets an instance of T and if it is decorated with @inject or any of the __init__ arguments are wrapped with Inject[...] the dependencies will be injected
        """
        return self._injector.get(interface=interface, scope=scope)

    def call_with_injection(
        self,
        callable: Callable[..., T],
        self_: Any = None,
        args: Optional[Any] = None,
        kwargs: Optional[Any] = None,
    ) -> T:
        """
        Calls the function and injects the parameters if the function is decorated with @inject or any of the arguments are wrapped with Inject[..]
        """
        args = args if args is not None else ()
        kwargs = kwargs if kwargs is not None else {}
        return self._injector.call_with_injection(callable, self_, args, kwargs)


__default_registry: Optional[Registry] = None


def set_default_registry(registry: Registry):
    global __default_registry
    __default_registry = registry


def get_default_registry() -> Registry:
    if __default_registry is None:
        set_default_registry(Registry())
    return __default_registry



from .loggers import logger

# from .helpers import *
from .registry import Registry, set_default_registry, get_default_registry
from .decorators import inject_from, inject, injectable
from .service_config import ServiceConfig
from injector import singleton, provider


# .fastapi moved to injectr
# dont'export it here so we can move it to a separate package later
# from .fastapi import FastApiRegistryBuilder



import logging

from src.fastinject import Registry, inject_from
from test.objects_for_testing import service_configs


def test_can_call_with_injection():
    # 1. Create registry
    registry = Registry(service_configs=[service_configs.SCLogging, service_configs.SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_both(dbcon: service_configs.DatabaseConfig, _logger: logging.Logger):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"
        assert _logger is not None

    registry.call_with_injection(callable=inject_both)


def test_can_call_with_injection_additional_args():
    # 1. Create registry
    registry = Registry(service_configs=[service_configs.SCLogging, service_configs.SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_both(dbcon: service_configs.DatabaseConfig, _logger: logging.Logger, name: str):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"
        assert _logger is not None
        assert name == "some_name"

    registry.call_with_injection(callable=inject_both, kwargs={"name": "some_name"})



import logging
from typing import List, Optional

import pytest
from injector import singleton, provider

from src.fastinject import Registry, inject, injectable, ServiceConfig
from test.objects_for_testing.service_configs_injectable import ModuleLogging, ModuleDatabase
from test.objects_for_testing.registries import DummyRegistry
from test.objects_for_testing.services import DatabaseConfig
from test.objects_for_testing.services_injectable import NotInjectedService, TimeStamp


class NonRegisteredClass:
    pass


@pytest.fixture(scope="function")
def test_can_inject_with_autobind():
    # 1. Decorate functions to inject from (default) registry. Registry is created with @injectable on service
    @inject()
    def inject_logger_in_fn(_logger: logging.Logger):
        assert _logger is not None

    # 3. Call decorated functions
    inject_logger_in_fn()


@pytest.fixture(scope="function")
def test_sc_catch_error_in_getting_service_from_registry():
    # 1. Decorate functions to inject from (default) registry. Registry is created with @injectable on service
    @inject()
    def injected_fn(noop: NotInjectedService, ts: TimeStamp):
        pass

    # 2. Call decorated functions
    with pytest.raises(TypeError):
        # TimeStamp is correctly decorated with @injectable on service, NotInjectedService is not
        injected_fn()


def test_inject_none_if_error_in_getting_optional_service_from_registry():
    # 1. Create registry
    registry = DummyRegistry()
    assert registry is not None

    # 2. Decorate functions with registry to inject from
    @inject()
    def injected_fn(val: Optional[NonRegisteredClass] = None):
        assert val is None

    # 3. Call decorated functions
    injected_fn()


@pytest.fixture(scope="function")
def test_raises_on_injecting_unregisterd_required_object():
    # 2. Decorate functions with registry to inject from
    @inject()
    def inject_logger_in_fn(my_inst: NonRegisteredClass): ...

    # Should fail because it cannot inject requried SomeClass type
    with pytest.raises(TypeError):
        inject_logger_in_fn()


def test_returns_none_on_injecting_unregisterd_optional_object():
    # 2. Decorate functions with registry to inject from
    @inject()
    def inject_logger_in_fn(my_inst: Optional[NonRegisteredClass]):
        assert my_inst is None

    # Should fail because it cannot inject requried SomeClass type
    inject_logger_in_fn()


def test_raises_on_injecting_unregisterd_optional_object_when_not_inject_none():
    # 1. Decorate functions with registry to inject from
    @inject(inject_missing_optional_as_none=False)
    def inject_logger_in_fn(my_inst: Optional[NonRegisteredClass]):
        assert my_inst is None

    # Should fail because it cannot inject requried SomeClass type
    with pytest.raises(TypeError):
        inject_logger_in_fn()


def test_sc_can_inject_from_with_optional_dependency():
    # 1. Create registry
    registry = Registry(service_configs=[ModuleLogging, ModuleDatabase])
    assert registry is not None

    # 2. Decorate functions with registry to inject from
    @inject()
    def inject_logger_in_fn(_logger: Optional[logging.Logger]):
        assert _logger is not None

    @inject()
    def inject_dbconfig_in_fn(dbcon: DatabaseConfig):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"

    @inject()
    def inject_both(dbcon: DatabaseConfig, _logger: logging.Logger):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"
        assert _logger is not None

    # 3. Call decorated functions
    inject_logger_in_fn()
    inject_dbconfig_in_fn()
    inject_both()


@pytest.fixture(scope="function")
def test_can_inject_from_with_additional_args():
    @inject()
    def inject_logger_in_fn(_logger: logging.Logger, a: int, b: int = 1, c: Optional[int] = None):
        assert _logger is not None
        assert isinstance(a, int)
        assert isinstance(b, int)
        assert isinstance(c, int) or c is None

    # 3. Call decorated functions
    inject_logger_in_fn(a=5, b=4, c=4)
    inject_logger_in_fn(a=5, b=4, c=None)
    inject_logger_in_fn(a=5)
    with pytest.raises(TypeError):
        # a is required; shoudle be provided
        inject_logger_in_fn()
        inject_logger_in_fn(b=5)
        inject_logger_in_fn(c=5)


@pytest.fixture(scope="function")
def test_raises_typeerror_on_nonregistered_type_with_injectable():
    """Cannot get type form container that isn't registered; throws"""

    @inject()
    def inject_non_existent(_logger: List):
        assert _logger is None

    with pytest.raises(TypeError):
        inject_non_existent()


def test_can_inject_in_class():
    # 2. Decorate class with registry to inject from
    class MyClass:
        @inject()
        def __init__(self, _logger: logging.Logger):
            self.logger = _logger
            assert _logger is not None

    # 3. Create instance and call decorated function
    my_class = MyClass()
    assert my_class is not None


def test_raises_when_decorating_wrongtype_class_with_injectables():
    @injectable()
    class ModuleDatabase(ServiceConfig):
        # todo test wheteher only some functions within a class can be providers
        @provider
        def provide_config(self) -> DatabaseConfig:
            return DatabaseConfig(connection_string="file:memdb1?mode=memory&cache=shared3")



import logging
import time
from typing import List, Optional

import pytest

from src.fastinject import (
    inject_from,
    Registry,
)
from test.objects_for_testing import services
from test.objects_for_testing.service_configs import SCLogging, SCDatabase, SCNestedDependenciesSimple
from test.objects_for_testing.registries import DummyRegistry
from test.objects_for_testing.services import DatabaseConfig, TimeStampLogger


class NonRegisteredClass:
    pass


def test_can_inject_from():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_logger_in_fn(_logger: logging.Logger):
        assert _logger is not None

    @inject_from(registry=registry)
    def inject_dbconfig_in_fn(dbcon: DatabaseConfig):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"

    @inject_from(registry=registry)
    def inject_both(dbcon: DatabaseConfig, _logger: logging.Logger):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"
        assert _logger is not None

    # 3. Call decorated functions
    inject_logger_in_fn()
    inject_dbconfig_in_fn()
    inject_both()


@pytest.fixture(scope="function")
def test_catch_error_in_getting_service_from_registry():
    # 1. Create registry
    registry = DummyRegistry()

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def injected_fn(dbcon: DatabaseConfig, _logger: logging.Logger):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"
        assert _logger is not None

    # 3. Call decorated functions
    with pytest.raises(ValueError):
        injected_fn()


def test_inject_none_if_error_in_getting_optional_service_from_registry():
    # 1. Create registry
    registry = DummyRegistry()

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def injected_fn(dbcon: Optional[DatabaseConfig] = None):
        assert dbcon is None

    # 3. Call decorated functions
    injected_fn()


@pytest.fixture(scope="function")
def test_raises_on_injecting_unregisterd_required_object():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_logger_in_fn(my_inst: NonRegisteredClass): ...

    # Should fail because it cannot inject requried SomeClass type
    with pytest.raises(TypeError):
        inject_logger_in_fn()


def test_returns_none_on_injecting_unregisterd_optional_object():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_logger_in_fn(my_inst: Optional[NonRegisteredClass]):
        assert my_inst is None

    # Should fail because it cannot inject requried SomeClass type
    inject_logger_in_fn()


def test_raises_on_injecting_unregisterd_optional_object_when_not_inject_none():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry, inject_missing_optional_as_none=False)
    def inject_logger_in_fn(my_inst: Optional[NonRegisteredClass]):
        assert my_inst is None

    # Should fail because it cannot inject requried SomeClass type
    with pytest.raises(TypeError):
        inject_logger_in_fn()


def test_can_inject_from_nested_dependencies():
    # 1. Create registry
    registry = Registry(service_configs=[SCNestedDependenciesSimple])

    @inject_from(registry=registry)
    def inject_logger_in_fn(ts: services.TimeStamp, tslogger: TimeStampLogger):
        assert ts is not None
        assert tslogger is not None

    # 3. Call decorated functions
    inject_logger_in_fn()
    time.sleep(0.1)
    inject_logger_in_fn()
    time.sleep(0.2)


def test_can_inject_from_with_optional_dependency():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_logger_in_fn(_logger: Optional[logging.Logger]):
        assert _logger is not None

    @inject_from(registry=registry)
    def inject_dbconfig_in_fn(dbcon: DatabaseConfig):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"

    @inject_from(registry=registry)
    def inject_both(dbcon: DatabaseConfig, _logger: logging.Logger):
        assert dbcon is not None
        assert dbcon.connection_string == "file:memdb1?mode=memory&cache=shared3"
        assert _logger is not None

    # 3. Call decorated functions
    inject_logger_in_fn()
    inject_dbconfig_in_fn()
    inject_both()


@pytest.fixture(scope="function")
def test_can_inject_from_with_additional_args():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate functions with registry to inject from
    @inject_from(registry=registry)
    def inject_logger_in_fn(_logger: logging.Logger, a: int, b: int = 1, c: Optional[int] = None):
        assert _logger is not None
        assert isinstance(a, int)
        assert isinstance(b, int)
        assert isinstance(c, int) or c is None

    # 3. Call decorated functions
    inject_logger_in_fn(a=5, b=4, c=4)
    inject_logger_in_fn(a=5, b=4, c=None)
    inject_logger_in_fn(a=5)
    with pytest.raises(TypeError):
        # a is required; shoudle be provided
        inject_logger_in_fn()
        inject_logger_in_fn(b=5)
        inject_logger_in_fn(c=5)


@pytest.fixture(scope="function")
def test_raises_typeerror_on_nonregistered_type():
    """Cannot get type form container that isn't registered; throws"""

    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    @inject_from(registry=registry)
    def inject_non_existent(_logger: List):
        assert _logger is None

    with pytest.raises(TypeError):
        inject_non_existent()


def test_can_inject_in_class():
    # 1. Create registry
    registry = Registry(service_configs=[SCLogging, SCDatabase])

    # 2. Decorate class with registry to inject from
    class MyClass:
        @inject_from(registry=registry)
        def __init__(self, _logger: logging.Logger):
            self.logger = _logger
            assert _logger is not None

    # 3. Create instance and call decorated function
    my_class = MyClass()
    assert my_class is not None


